// Copyright 1994-2000 by Jon Dart.  All Rights Reserved.

#include "board.h"
#include "constant.h"
#include "bhash.h"
#include "util.h"
#include "bearing.h" 
#include "debug.h"
#include "boardio.h"
#include "boardsta.h"
#include <ctype.h>
#include <memory.h>
#include <assert.h>
#include <iostream>

using namespace std;

enum
{
   BoardSize = 64
};
enum
{
   RepListSize = 1024
};

void Board::set_secondary_vars()
{
   int index[2], i;

   index[Black] = index[White] = 0;

   my_Material[White].clear();
   my_Material[Black].clear();
   pawn_bits[White].clear();
   pawn_bits[Black].clear();
   knight_bits[White].clear();
   knight_bits[Black].clear();
   bishop_bits[White].clear();
   bishop_bits[Black].clear();
   rook_bits[White].clear();
   rook_bits[Black].clear();
   rooksqueens[White].clear();
   rooksqueens[Black].clear();
   bishopsqueens[White].clear();
   bishopsqueens[Black].clear();
   queen_bits[White].clear();
   queen_bits[Black].clear();
   occupied[White].clear();
   occupied[Black].clear();
   all_occupied.clear();
   occupiedR90.clear();
   occupiedDa1.clear();
   occupiedDa8.clear();
   for (i=0;i<BoardSize;i++)
   {
      Square sq(i);
      if (my_Contents[sq] != EmptyPiece)
      {
         const Piece piece = my_Contents[sq];
         const ColorType color = PieceColor(piece);
         occupied[color].set(sq);
         all_occupied.set(sq);
         occupiedR90.setR90(sq);
         occupiedDa1.setDa1(sq);
         occupiedDa8.setDa8(sq);
         my_Material[color].add_piece(TypeOfPiece(piece));
         switch (TypeOfPiece(piece))
         {
         case King:
            my_KingPos[color] = sq;
            break;
         case Pawn:
            pawn_bits[color].set(sq);
            break;
         case Knight:
            knight_bits[color].set(sq);
            break;
         case Bishop:
            bishop_bits[color].set(sq);
            bishopsqueens[color].set(sq);
            break;
         case Rook:
            rook_bits[color].set(sq);
            rooksqueens[color].set(sq);
            break;
         case Queen:
            queen_bits[color].set(sq);
            rooksqueens[color].set(sq);
            bishopsqueens[color].set(sq);
            break;
         default:
            break;
         }
      }
   }
   hashCode = BoardHash::hashCode(*this);
   m_pawn_hash = BoardHash::pawnHash(*this);
}

void Board::Reset()
{
   static PieceType pieces[] =
   {
      Rook,
      Knight,
      Bishop,
      Queen,
      King,
      Bishop,
      Knight,
      Rook
   };

   side = White;
   my_CheckStatus = Unknown;
   for (int i=0;i<BoardSize;i++)
   {
      const Square sq(i);
      if (Rank(sq,White) == 1)
         my_Contents[sq] = MakeWhitePiece( pieces[File(sq)-1]);
      else if (Rank(sq,Black) == 1)
         my_Contents[sq] = MakeBlackPiece( pieces[File(sq)-1]);
      else if (Rank(sq,White) == 2)
         my_Contents[sq] = whitePawn;
      else if (Rank(sq,Black) == 2)
         my_Contents[sq] = blackPawn;
      else
         my_Contents[sq] = EmptyPiece;
   }
   my_EnPassantSq = InvalidSquare;
   my_CastleStatus[White] = my_CastleStatus[Black] = CanCastleEitherSide;
   if (rep_list)
      delete [] rep_list;
   rep_list = rep_list_head = new hash_t[RepListSize];
   move_count = 0;
   set_secondary_vars();
}

Board::Board()
: rep_list(NULL)
{
   Reset();
}

Board::Board(const Board &b)
{
   side = b.side;
   my_CheckStatus = b.my_CheckStatus;
   memcpy(&my_Contents,&b.my_Contents,BoardSize*sizeof(Square));
   my_EnPassantSq = b.my_EnPassantSq;
   my_CastleStatus[White] = b.my_CastleStatus[White];
   my_CastleStatus[Black] = b.my_CastleStatus[Black];

   // Copy the repetition table
   move_count = b.move_count;
   rep_list = new hash_t[RepListSize];
   int rep_entries = b.rep_list_head - b.rep_list;
   for (int i = 0; i < rep_entries; i++)
      rep_list[i] = b.rep_list[i];
   rep_list_head = rep_list + rep_entries;

   set_secondary_vars();
}

Board &Board::operator =(const Board &b)
{
   if (&b != this)
   {
      side = b.side;
      my_CheckStatus = b.my_CheckStatus;
      memcpy(&my_Contents,&b.my_Contents,BoardSize*sizeof(Square));
      my_EnPassantSq = b.my_EnPassantSq;
      my_CastleStatus[White] = b.my_CastleStatus[White];
      my_CastleStatus[Black] = b.my_CastleStatus[Black];
      set_secondary_vars();

      // Copy the repetition table
      move_count = b.move_count;
      delete [] rep_list;
      rep_list = new hash_t[RepListSize];
      int rep_entries = b.rep_list_head - b.rep_list;
      for (int i = 0; i < rep_entries; i++)
         rep_list[i] = b.rep_list[i];
      rep_list_head = rep_list + rep_entries;
   }
   return *this;
}

Board::~Board()
{
   delete [] rep_list;
}

#ifdef _DEBUG
const Piece &Board::operator[]( const Square sq ) const
{
   ASSERT(OnBoard(sq));
   return my_Contents[sq];
}
#endif

void Board::UpdateCastleStatus( const ColorType side, 
                                const Square sq )
// after a move of or capture of the rook on 'sq', update castle status
// for 'side'
{
   CastleType cs = CastleStatus(side);
   if ((cs != CastledKSide) && (cs != CastledQSide))
   {
      if (File(sq) == 1 && Rank(sq,side)==1) // Queen Rook moved or captured
      {
         if (cs == CanCastleEitherSide)
            my_CastleStatus[side] = CanCastleKSide;
         else if (cs == CanCastleQSide)
            my_CastleStatus[side] = CantCastleEitherSide;
         if (side == White)
         {
           hashCode ^= w_castle_status[(int)cs];
           hashCode ^= w_castle_status[(int)my_CastleStatus[White]];
         }
         else
         {
           hashCode ^= b_castle_status[(int)cs];
           hashCode ^= b_castle_status[(int)my_CastleStatus[Black]];
         }
      }
      else if (File(sq) == 8 && Rank(sq,side)==1) // King Rook moved or captured
      {
         if (cs == CanCastleEitherSide)
            my_CastleStatus[side] = CanCastleQSide;
         else if (cs == CanCastleKSide)
            my_CastleStatus[side] = CantCastleEitherSide;
         if (side == White)
         {
           hashCode ^= w_castle_status[(int)cs];
           hashCode ^= w_castle_status[(int)my_CastleStatus[White]];
         }
         else
         {
           hashCode ^= b_castle_status[(int)cs];
           hashCode ^= b_castle_status[(int)my_CastleStatus[Black]];
         }
      }
   }
}

int Board::Between( const Square &sq1, const Square &sq2, 
                    Square *squares) const
{
   int offset = Bearing::Directions[sq1][sq2];
   if (offset == 0) return 0;
   Square sq = sq1;
   int n = 0;
   for (;;)
   {
      sq += offset;
      if (sq == sq2)
         break;
      ASSERT(OnBoard(sq));
      squares[n++] = sq;
   }
   return n;
}


int Board::Clear( const Square &sq1, const Square &sq2 ) const
{
   int offset = Bearing::Directions[sq1][sq2];
   if (offset == 0) return 0;
   Square sq = sq1;
   for (;;)
   {
      sq += offset;
      if (sq == sq2)
         break;
      ASSERT(OnBoard(sq));
      if (my_Contents[sq] != EmptyPiece)
         return 0;
   }
   return 1;
}

#define Xor(h,sq,piece) h ^= hash_codes[sq][(int)piece]

void Board::DoNull()
{
   my_CheckStatus = Unknown;
   move_count++;
   if (my_EnPassantSq != InvalidSquare)
   {
       hashCode ^= ep_codes[my_EnPassantSq];
       hashCode ^= ep_codes[0];
   }
   my_EnPassantSq = InvalidSquare;
   side = OppositeSide();
   if (Side() == Black)
       hashCode |= (hash_t)1;
   else
       hashCode &= (hash_t)~1;
   *rep_list_head++ = hashCode;
   ASSERT(rep_list_head-rep_list < (int)RepListSize);
   ASSERT(hashCode == BoardHash::hashCode(*this));
}

void Board::DoMove( const Move &move )
{
#ifdef _DEBUG
   Piece targetPiece;
#endif
   my_CheckStatus = Unknown;
   move_count++;
   if (my_EnPassantSq != InvalidSquare)
   {
       hashCode ^= ep_codes[my_EnPassantSq];
       hashCode ^= ep_codes[0];
   }
   int old_epsq = my_EnPassantSq;
   my_EnPassantSq = InvalidSquare;

   ASSERT(!IsEmptyPiece(PieceMoved(move)));

   const Square start = StartSquare(move);
   const Square dest = DestSquare(move);
   const SpecialType moveType = TypeOfMove(move);
   if (side == White)
   {
      if (moveType == KCastle)
      {
         move_count = 0;

         // update the hash code
         const Square kp = KingPos(White);
         Xor(hashCode, kp+3, whiteRook);
         Xor(hashCode, kp, whiteKing);
         Xor(hashCode, kp+1, whiteRook);
         Xor(hashCode, kp+2, whiteKing);
         hashCode ^= w_castle_status[(int)my_CastleStatus[White]];
         hashCode ^= w_castle_status[(int)CastledKSide];

         const int newkp = kp + 2;
         my_KingPos[White] = newkp;
         my_CastleStatus[White] = CastledKSide;
         // find old square of rook
         Square oldrooksq = kp + 3;
         Square newrooksq = kp + 1;
         my_Contents[kp] = my_Contents[oldrooksq] = EmptyPiece;
         my_Contents[newrooksq] = whiteRook;
         my_Contents[newkp] = whiteKing;
         rook_bits[White].clear(oldrooksq);
         rooksqueens[White].clear(oldrooksq);
         rook_bits[White].set(newrooksq);
         rooksqueens[White].set(newrooksq);
         clearAll(White,kp);
         clearAll(White,oldrooksq);
         setAll(White,newkp);
         setAll(White,newrooksq);
      }
      else if (moveType == QCastle)
      {
         move_count = 0;

         // update the hash code
         const Square kp = KingPos(White);
         Xor(hashCode, kp-4, whiteRook);
         Xor(hashCode, kp, whiteKing);
         Xor(hashCode, kp-1, whiteRook);
         Xor(hashCode, kp-2, whiteKing);
         hashCode ^= w_castle_status[(int)my_CastleStatus[White]];
         hashCode ^= w_castle_status[(int)CastledQSide];

         const int newkp = kp - 2;
         my_KingPos[White] = newkp;
         my_CastleStatus[White] = CastledQSide;
         // find old square of rook
         Square oldrooksq = kp - 4;
         Square newrooksq = kp - 1;
         my_Contents[kp] = my_Contents[oldrooksq] = Piece();
         my_Contents[newrooksq] = whiteRook;
         my_Contents[kp-2] = whiteKing;
         rook_bits[White].clear(oldrooksq);
         rooksqueens[White].clear(oldrooksq);
         rook_bits[White].set(newrooksq);
         rooksqueens[White].set(newrooksq);
         clearAll(White,kp);
         clearAll(White,oldrooksq);
         setAll(White,newkp);
         setAll(White,newrooksq);
      }
      else // not castling
      {
         ASSERT(my_Contents[start] != EmptyPiece);
         const Bitmap bits(Bitmap::mask[start] |
                           Bitmap::mask[dest]);
         Square target = dest; // where we captured
#ifdef _DEBUG
         targetPiece = my_Contents[target];
#endif
         switch (PieceMoved(move))
         {
         case Pawn:
            move_count = 0;
            switch (moveType)
            {
            case EnPassant:
               // update hash code
               Xor(hashCode, start, whitePawn);
               Xor(hashCode, dest, whitePawn);
               Xor(m_pawn_hash, start, whitePawn);
               Xor(m_pawn_hash, dest, whitePawn);
               target = old_epsq;
#ifdef _DEBUG
               targetPiece = my_Contents[target];
#endif
               my_Contents[dest] = whitePawn;
               pawn_bits[White].set(dest);
               break;
            case Promotion:
               // update hash code
               Xor(hashCode, start, whitePawn);
               Xor(hashCode, dest, MakeWhitePiece(PromoteTo(move)));
               Xor(m_pawn_hash, start, whitePawn);
               my_Contents[dest] = MakeWhitePiece(PromoteTo(move));
               my_Material[White].remove_pawn();
               my_Material[White].add_piece(PromoteTo(move));
               switch (PromoteTo(move))
               {
               case Knight:
                  knight_bits[White].set(dest);
                  break;
               case Bishop:
                  bishop_bits[White].set(dest);
                  bishopsqueens[White].set(dest);
                  break;
               case Rook:
                  rook_bits[White].set(dest);
                  rooksqueens[White].set(dest);
                  break;
               case Queen:
                  queen_bits[White].set(dest);
                  bishopsqueens[White].set(dest);
                  rooksqueens[White].set(dest);
                  break;
               default:
                  break;
               }
               break;
            default:
               Xor(hashCode, start, whitePawn );
               Xor(hashCode, dest, whitePawn );
               Xor(m_pawn_hash, start, whitePawn);
               Xor(m_pawn_hash, dest, whitePawn);
               my_Contents[dest] = whitePawn;
               if (start - dest == 16) // 2-square pawn advance
               {
                  if (TEST_MASK(Bearing::ep_mask[File(dest)-1][(int)White],pawn_bits[Black])) {
                    my_EnPassantSq = dest;
                    hashCode ^= ep_codes[0];
                    hashCode ^= ep_codes[dest];
                  }
               }
               pawn_bits[White].set(dest);
               break;
            }
            pawn_bits[White].clear(start);
            break;
         case Knight:
            Xor(hashCode, start, whiteKnight);
            Xor(hashCode, dest, whiteKnight);
            my_Contents[dest] = whiteKnight;
            knight_bits[White].setClear(bits);
            break;
         case Bishop:
            Xor(hashCode, start, whiteBishop);
            Xor(hashCode, dest, whiteBishop);
            my_Contents[dest] = whiteBishop;
            bishop_bits[White].setClear(bits);
            bishopsqueens[White].setClear(bits);
            break;
         case Rook:
            Xor(hashCode, start, whiteRook );
            Xor(hashCode, dest, whiteRook );
            my_Contents[dest] = whiteRook;
            rook_bits[White].setClear(bits);
            rooksqueens[White].setClear(bits);
            UpdateCastleStatus(side,start);
            break;
         case Queen:
            Xor(hashCode, start, whiteQueen);
            Xor(hashCode, dest, whiteQueen);
            my_Contents[dest] = whiteQueen;
            queen_bits[White].setClear(bits);
            rooksqueens[White].setClear(bits);
            bishopsqueens[White].setClear(bits);
            break;
         case King:
            Xor(hashCode, start, whiteKing );
            Xor(hashCode, dest, whiteKing );
            my_Contents[dest] = whiteKing;
            my_KingPos[White] = dest;
            if ((CastleStatus(White) != CastledQSide) &&
                (CastleStatus(White) != CastledKSide))
            {
               hashCode ^= w_castle_status[(int)CastleStatus(White)];
               hashCode ^= w_castle_status[(int)CantCastleEitherSide];
               my_CastleStatus[White] = CantCastleEitherSide;
            }
            break;
         }
         my_Contents[start] = EmptyPiece;
         if (Capture(move) != EmptyPiece) 
         {
            move_count = 0;
            ASSERT(target != InvalidSquare);
#ifdef _DEBUG
            ASSERT(targetPiece != EmptyPiece &&
                   PieceColor(targetPiece) == Black);
#endif
            occupied[Black].clear(target);
            Piece cap = MakeBlackPiece(Capture(move));
            Xor(hashCode, target, cap);
            switch (Capture(move))
            {
            case Pawn:
               pawn_bits[Black].clear(target);
               Xor(m_pawn_hash, target, cap);
               if (moveType == EnPassant)
               {
                  my_Contents[target] = EmptyPiece;
                  clearAll(Black,target);
               }
               my_Material[Black].remove_pawn();
               break;
            case Rook:
               UpdateCastleStatus(Black,dest);
               rook_bits[Black].clear(target);
               rooksqueens[Black].clear(target);
               my_Material[Black].remove_piece(Rook);
               break;
            case Knight:
               knight_bits[Black].clear(target);
               my_Material[Black].remove_piece(Knight);
               break;
            case Bishop:
               bishop_bits[Black].clear(target);
               bishopsqueens[Black].clear(target);
               my_Material[Black].remove_piece(Bishop);
               break;
            case Queen:
               queen_bits[Black].clear(target);
               rooksqueens[Black].clear(target);
               bishopsqueens[Black].clear(target);
               my_Material[Black].remove_piece(Queen);
               break;
            case King:
               ASSERT(0);
               my_KingPos[Black] = InvalidSquare;
               my_CastleStatus[Black] = CantCastleEitherSide;
               my_Material[Black].remove_piece(King);
               break;
            default: 
               break;
            }
         }
      }
      setAll(White,dest);
      clearAll(White,start);

   }
   else // side == Black
   {
      if (moveType == KCastle)
      {
         move_count = 0;
         const Square kp = KingPos(Black);

         // update the hash code
         Xor(hashCode, kp+3, blackRook);
         Xor(hashCode, kp, blackKing);
         Xor(hashCode, kp+1, blackRook);
         Xor(hashCode, kp+2, blackKing);
         hashCode ^= b_castle_status[(int)my_CastleStatus[Black]];
         hashCode ^= b_castle_status[(int)CastledKSide];

         const int newkp = kp + 2;
         my_KingPos[Black] = newkp;
         my_CastleStatus[Black] = CastledKSide;
         // find old square of rook
         Square oldrooksq = kp + 3;
         Square newrooksq = kp + 1;
         my_Contents[kp] = my_Contents[oldrooksq] = EmptyPiece;
         my_Contents[newrooksq] = blackRook;
         my_Contents[kp+2] = blackKing;
         rook_bits[Black].clear(oldrooksq);
         rooksqueens[Black].clear(oldrooksq);
         rook_bits[Black].set(newrooksq);
         rooksqueens[Black].set(newrooksq);
         clearAll(Black,kp);
         clearAll(Black,oldrooksq);
         setAll(Black,newkp);
         setAll(Black,newrooksq);
      }
      else if (moveType == QCastle)
      {
         move_count = 0;
         const Square kp = KingPos(Black);

         // update the hash code
         Xor(hashCode, kp-4, blackRook);
         Xor(hashCode, kp, blackKing);
         Xor(hashCode, kp-1, blackRook);
         Xor(hashCode, kp-2, blackKing);
         hashCode ^= b_castle_status[(int)my_CastleStatus[Black]];
         hashCode ^= b_castle_status[(int)CastledQSide];

         const int newkp = kp - 2;
         my_KingPos[Black] = newkp;
         my_CastleStatus[Black] = CastledQSide;
         // find old square of rook
         Square oldrooksq = kp - 4;
         Square newrooksq = kp - 1;
         my_Contents[kp] = my_Contents[oldrooksq] = EmptyPiece;
         my_Contents[newrooksq] = blackRook;
         my_Contents[kp-2] = blackKing;
         rook_bits[Black].clear(oldrooksq);
         rooksqueens[Black].clear(oldrooksq);
         rook_bits[Black].set(newrooksq);
         rooksqueens[Black].set(newrooksq);
         clearAll(Black,kp);
         clearAll(Black,oldrooksq);
         setAll(Black,newkp);
         setAll(Black,newrooksq);
      }
      else // not castling
      {
  	 ASSERT(my_Contents[start] != EmptyPiece);
         const Bitmap bits(Bitmap::mask[start] |
                           Bitmap::mask[dest]);
         Square target = dest; // where we captured
#ifdef _DEBUG
         targetPiece = my_Contents[target];
#endif
         switch (PieceMoved(move))
         {
         case Pawn:
            move_count = 0;
            switch (moveType)
            {
            case EnPassant:
               // update hash code
               Xor(hashCode, start, blackPawn);
               Xor(hashCode, dest, blackPawn);
               Xor(m_pawn_hash, start, blackPawn);
               Xor(m_pawn_hash, dest, blackPawn);
               target = old_epsq;
#ifdef _DEBUG
               targetPiece = my_Contents[target];
#endif
               my_Contents[dest] = blackPawn;
               pawn_bits[Black].set(dest);
               break;
            case Promotion:
               // update hash code
               Xor(hashCode, start, blackPawn);
               Xor(hashCode, dest, MakeBlackPiece(PromoteTo(move)));
               Xor(m_pawn_hash, start, blackPawn);
               my_Contents[dest] = MakeBlackPiece(PromoteTo(move));
               my_Material[Black].remove_pawn();
               my_Material[Black].add_piece(PromoteTo(move));
               switch (PromoteTo(move))
               {
               case Knight:
                  knight_bits[Black].set(dest);
                  break;
               case Bishop:
                  bishop_bits[Black].set(dest);
                  bishopsqueens[Black].set(dest);
                  break;
               case Rook:
                  rook_bits[Black].set(dest);
                  rooksqueens[Black].set(dest);
                  break;
               case Queen:
                  queen_bits[Black].set(dest);
                  bishopsqueens[Black].set(dest);
                  rooksqueens[Black].set(dest);
                  break;
               default:
                  break;
               }
               break;
            default:
               Xor(hashCode, start, blackPawn );
               Xor(hashCode, dest, blackPawn );
               Xor(m_pawn_hash, start, blackPawn);
               Xor(m_pawn_hash, dest, blackPawn);
               my_Contents[dest] = blackPawn;
               if (dest - start == 16) // 2-square pawn advance
               { 
                  if (TEST_MASK(Bearing::ep_mask[File(dest)-1][(int)Black],pawn_bits[White])) {
                    my_EnPassantSq = dest;
                    hashCode ^= ep_codes[0];
                    hashCode ^= ep_codes[dest];
                  }
               }
               pawn_bits[Black].set(dest);
               break;
            }
            pawn_bits[Black].clear(start);
            break;
         case Knight:
            Xor(hashCode, start, blackKnight);
            Xor(hashCode, dest, blackKnight);
            my_Contents[dest] = blackKnight;
            knight_bits[Black].setClear(bits);
            break;
         case Bishop:
            Xor(hashCode, start, blackBishop);
            Xor(hashCode, dest, blackBishop);
            my_Contents[dest] = blackBishop;
            bishop_bits[Black].setClear(bits);
            bishopsqueens[Black].setClear(bits);
            break;
         case Rook:
            Xor(hashCode, start, blackRook );
            Xor(hashCode, dest, blackRook );
            my_Contents[dest] = blackRook;
            rook_bits[Black].setClear(bits);
            rooksqueens[Black].setClear(bits);
            UpdateCastleStatus(side,start);
            break;
         case Queen:
            Xor(hashCode, start, blackQueen);
            Xor(hashCode, dest, blackQueen);
            my_Contents[dest] = blackQueen;
            queen_bits[Black].setClear(bits);
            rooksqueens[Black].setClear(bits);
            bishopsqueens[Black].setClear(bits);
            break;
         case King:
            Xor(hashCode, start, blackKing );
            Xor(hashCode, dest, blackKing );
            my_Contents[dest] = blackKing;
            my_KingPos[Black] = dest;
            if ((CastleStatus(Black) != CastledQSide) &&
                (CastleStatus(Black) != CastledKSide))
            {
               hashCode ^= b_castle_status[(int)CastleStatus(Black)];
               hashCode ^= b_castle_status[(int)CantCastleEitherSide];
               my_CastleStatus[Black] = CantCastleEitherSide;
            }
            break;
         }
         my_Contents[start] = EmptyPiece;
         if (Capture(move) != EmptyPiece)
         {
            move_count = 0;
            ASSERT(target != InvalidSquare);
#ifdef _DEBUG
            ASSERT(targetPiece != EmptyPiece  &&
                   PieceColor(targetPiece) == White);
#endif
            occupied[White].clear(target);
            Piece cap = MakeWhitePiece(Capture(move));
            Xor(hashCode, target, cap);
            switch (Capture(move))
            {
            case Pawn:
               pawn_bits[White].clear(target);
               Xor(m_pawn_hash, target, cap);
               if (moveType == EnPassant)
               {
                  my_Contents[target] = EmptyPiece;
                  clearAll(White,target);
               }
               my_Material[White].remove_pawn();
               break;
            case Rook:
               UpdateCastleStatus(White,dest);
               rook_bits[White].clear(target);
               rooksqueens[White].clear(target);
               my_Material[White].remove_piece(Rook);
               break;
            case Knight:
               knight_bits[White].clear(target);
               my_Material[White].remove_piece(Knight);
               break;
            case Bishop:
               bishop_bits[White].clear(target);
               bishopsqueens[White].clear(target);
               my_Material[White].remove_piece(Bishop);
               break;
            case Queen:
               queen_bits[White].clear(target);
               rooksqueens[White].clear(target);
               bishopsqueens[White].clear(target);
               my_Material[White].remove_piece(Queen);
               break;
            case King:
               ASSERT(0);
               my_KingPos[White] = InvalidSquare;
               my_CastleStatus[White] = CantCastleEitherSide;
               my_Material[White].remove_piece(King);
               break;
            default: 
               break;
            }
         }
         setAll(Black,dest);
         clearAll(Black,start);
      }
   }

   ASSERT(rep_list_head-rep_list < (int)RepListSize);
   if (Side() == White)
      hashCode |= (hash_t)1;
   else
      hashCode &= (hash_t)~1;
   *rep_list_head++ = hashCode;
   ASSERT(pawn_hash() == BoardHash::pawnHash(*this));
   side = OppositeSide();
   ASSERT(hashCode == BoardHash::hashCode(*this));
   all_occupied = occupied[White];
   all_occupied.Or(occupied[Black]);
#ifdef _DEBUG
   // verify correct updating of bitmaps:
   Board copy(*this);
   copy.set_secondary_vars();
   ASSERT(pawn_bits[White] == copy.pawn_bits[White]);
   ASSERT(knight_bits[White] == copy.knight_bits[White]);
   ASSERT(bishop_bits[White] == copy.bishop_bits[White]);
   ASSERT(rook_bits[White] == copy.rook_bits[White]);
   ASSERT(queen_bits[White] == copy.queen_bits[White]);
   ASSERT(bishopsqueens[White] == copy.bishopsqueens[White]);
   ASSERT(rooksqueens[White] == copy.rooksqueens[White]);
   ASSERT(occupied[White] == copy.occupied[White]);
   ASSERT(occupiedR90 == copy.occupiedR90);
   ASSERT(occupiedDa1 == copy.occupiedDa1);
   ASSERT(occupiedDa8 == copy.occupiedDa8);

   ASSERT(pawn_bits[Black] == copy.pawn_bits[Black]);
   ASSERT(knight_bits[Black] == copy.knight_bits[Black]);
   ASSERT(bishop_bits[Black] == copy.bishop_bits[Black]);
   ASSERT(rook_bits[Black] == copy.rook_bits[Black]);
   ASSERT(queen_bits[Black] == copy.queen_bits[Black]);
   ASSERT(bishopsqueens[Black] == copy.bishopsqueens[Black]);
   ASSERT(rooksqueens[Black] == copy.rooksqueens[Black]);
   ASSERT(occupied[Black] == copy.occupied[Black]);
   ASSERT(occupiedR90 == copy.occupiedR90);
   ASSERT(my_Contents[my_KingPos[White]]==whiteKing);
   ASSERT(my_Contents[my_KingPos[Black]]==blackKing);
#endif     
}

void Board::DoMoveQ( const Move &move )
{
#ifdef _DEBUG
   Piece targetPiece;
#endif
   my_CheckStatus = Unknown;
   move_count++;

   ASSERT(!IsNull(move));
   ASSERT(!IsEmptyPiece(PieceMoved(move)));

   Square old_epsq = my_EnPassantSq;
   my_EnPassantSq = InvalidSquare;

   const Square start = StartSquare(move);
   const Square dest = DestSquare(move);
   const SpecialType moveType = TypeOfMove(move);
   if (side == White)
   {
      if (moveType == KCastle)
      {
         move_count = 0;

         // update the hash code
         const Square kp = KingPos(White);
         const int newkp = kp + 2;
         my_KingPos[White] = newkp;
         my_CastleStatus[White] = CastledKSide;
         // find old square of rook
         Square oldrooksq = kp + 3;
         Square newrooksq = kp + 1;
         my_Contents[kp] = my_Contents[oldrooksq] = EmptyPiece;
         my_Contents[newrooksq] = whiteRook;
         my_Contents[newkp] = whiteKing;
         rook_bits[White].clear(oldrooksq);
         rooksqueens[White].clear(oldrooksq);
         rook_bits[White].set(newrooksq);
         rooksqueens[White].set(newrooksq);
         clearAll(White,kp);
         clearAll(White,oldrooksq);
         setAll(White,newkp);
         setAll(White,newrooksq);
      }
      else if (moveType == QCastle)
      {
         move_count = 0;

         const Square kp = KingPos(White);
         const int newkp = kp - 2;
         my_KingPos[White] = newkp;
         my_CastleStatus[White] = CastledQSide;
         // find old square of rook
         Square oldrooksq = kp - 4;
         Square newrooksq = kp - 1;
         my_Contents[kp] = my_Contents[oldrooksq] = Piece();
         my_Contents[newrooksq] = whiteRook;
         my_Contents[kp-2] = whiteKing;
         rook_bits[White].clear(oldrooksq);
         rooksqueens[White].clear(oldrooksq);
         rook_bits[White].set(newrooksq);
         rooksqueens[White].set(newrooksq);
         clearAll(White,kp);
         clearAll(White,oldrooksq);
         setAll(White,newkp);
         setAll(White,newrooksq);
      }
      else // not castling
      {
         const Bitmap bits(Bitmap::mask[start] |
                           Bitmap::mask[dest]);
         Square target = dest; // where we captured
#ifdef _DEBUG
         targetPiece = my_Contents[target];
#endif
         switch (PieceMoved(move))
         {
         case Pawn:
            move_count = 0;
            switch (moveType)
            {
            case EnPassant:
               Xor(m_pawn_hash, start, whitePawn);
               Xor(m_pawn_hash, dest, whitePawn);
               target = old_epsq;
#ifdef _DEBUG
               targetPiece = my_Contents[target];
#endif
               my_Contents[dest] = whitePawn;
               pawn_bits[White].set(dest);
               break;
            case Promotion:
               // update hash code
               Xor(m_pawn_hash, start, whitePawn);
               my_Contents[dest] = MakeWhitePiece(PromoteTo(move));
               my_Material[White].remove_pawn();
               my_Material[White].add_piece(PromoteTo(move));
               switch (PromoteTo(move))
               {
               case Knight:
                  knight_bits[White].set(dest);
                  break;
               case Bishop:
                  bishop_bits[White].set(dest);
                  bishopsqueens[White].set(dest);
                  break;
               case Rook:
                  rook_bits[White].set(dest);
                  rooksqueens[White].set(dest);
                  break;
               case Queen:
                  queen_bits[White].set(dest);
                  bishopsqueens[White].set(dest);
                  rooksqueens[White].set(dest);
                  break;
               default:
                  break;
               }
               break;
            default:
               Xor(m_pawn_hash, start, whitePawn);
               Xor(m_pawn_hash, dest, whitePawn);
               my_Contents[dest] = whitePawn;
               if (start - dest == 16) // 2-square pawn advance
               {
		 if (TEST_MASK(Bearing::ep_mask[File(dest)-1][(int)White],pawn_bits[Black])) {
                    my_EnPassantSq = dest;
		 }
               }
               pawn_bits[White].set(dest);
               break;
            }
            pawn_bits[White].clear(start);
            break;
         case Knight:
            my_Contents[dest] = whiteKnight;
            knight_bits[White].setClear(bits);
            break;
         case Bishop:
            my_Contents[dest] = whiteBishop;
            bishop_bits[White].setClear(bits);
            bishopsqueens[White].setClear(bits);
            break;
         case Rook:
            my_Contents[dest] = whiteRook;
            rook_bits[White].setClear(bits);
            rooksqueens[White].setClear(bits);
            UpdateCastleStatus(White,start);
            break;
         case Queen:
            my_Contents[dest] = whiteQueen;
            queen_bits[White].setClear(bits);
            rooksqueens[White].setClear(bits);
            bishopsqueens[White].setClear(bits);
            break;
         case King:
            my_Contents[dest] = whiteKing;
            my_KingPos[White] = dest;
            if ((CastleStatus(White) != CastledQSide) &&
                (CastleStatus(White) != CastledKSide))
            {
               my_CastleStatus[White] = CantCastleEitherSide;
            }
            break;
         }
         my_Contents[start] = EmptyPiece;
         if (Capture(move) != EmptyPiece)
         {
            move_count = 0;
            ASSERT(target != InvalidSquare);
#ifdef _DEBUG
            ASSERT(targetPiece != EmptyPiece &&
                   PieceColor(targetPiece) == Black);
#endif
            occupied[Black].clear(target);
            Piece cap = MakeBlackPiece(Capture(move));
            switch (Capture(move))
            {
            case Pawn:
               pawn_bits[Black].clear(target);
               Xor(m_pawn_hash, target, cap);
               if (moveType == EnPassant)
               {
                  my_Contents[target] = EmptyPiece;
                  clearAll(Black,target);
               }
               my_Material[Black].remove_pawn();
               break;
            case Rook:
               UpdateCastleStatus(Black,dest);
               rook_bits[Black].clear(target);
               rooksqueens[Black].clear(target);
               my_Material[Black].remove_piece(Rook);
               break;
            case Knight:
               knight_bits[Black].clear(target);
               my_Material[Black].remove_piece(Knight);
               break;
            case Bishop:
               bishop_bits[Black].clear(target);
               bishopsqueens[Black].clear(target);
               my_Material[Black].remove_piece(Bishop);
               break;
            case Queen:
               queen_bits[Black].clear(target);
               rooksqueens[Black].clear(target);
               bishopsqueens[Black].clear(target);
               my_Material[Black].remove_piece(Queen);
               break;
            case King:
               ASSERT(0);
               my_KingPos[Black] = InvalidSquare;
               my_CastleStatus[Black] = CantCastleEitherSide;
               my_Material[Black].remove_piece(King);
               break;
            default: 
               break;
            }
         }
         setAll(White,dest);
         clearAll(White,start);
      }

   }
   else // side == Black
   {
      if (moveType == KCastle)
      {
         move_count = 0;
         const Square kp = KingPos(Black);
         const int newkp = kp + 2;
         my_KingPos[Black] = newkp;
         my_CastleStatus[Black] = CastledKSide;
         // find old square of rook
         Square oldrooksq = kp + 3;
         Square newrooksq = kp + 1;
         my_Contents[kp] = my_Contents[oldrooksq] = EmptyPiece;
         my_Contents[newrooksq] = blackRook;
         my_Contents[kp+2] = blackKing;
         rook_bits[Black].clear(oldrooksq);
         rooksqueens[Black].clear(oldrooksq);
         rook_bits[Black].set(newrooksq);
         rooksqueens[Black].set(newrooksq);
         clearAll(Black,kp);
         clearAll(Black,oldrooksq);
         setAll(Black,newkp);
         setAll(Black,newrooksq);
      }
      else if (moveType == QCastle)
      {
         move_count = 0;
         const Square kp = KingPos(Black);
         const int newkp = kp - 2;
         my_KingPos[Black] = newkp;
         my_CastleStatus[Black] = CastledQSide;
         // find old square of rook
         Square oldrooksq = kp - 4;
         Square newrooksq = kp - 1;
         my_Contents[kp] = my_Contents[oldrooksq] = EmptyPiece;
         my_Contents[newrooksq] = blackRook;
         my_Contents[kp-2] = blackKing;
         rook_bits[Black].clear(oldrooksq);
         rooksqueens[Black].clear(oldrooksq);
         rook_bits[Black].set(newrooksq);
         rooksqueens[Black].set(newrooksq);
         clearAll(Black,kp);
         clearAll(Black,oldrooksq);
         setAll(Black,newkp);
         setAll(Black,newrooksq);
      }
      else // not castling
      {
         const Bitmap bits(Bitmap::mask[start] |
                           Bitmap::mask[dest]);
         Square target = dest; // where we captured
#ifdef _DEBUG
         Piece targetPiece = my_Contents[target];
#endif
         switch (PieceMoved(move))
         {
         case Pawn:
            move_count = 0;
            switch (moveType)
            {
            case EnPassant:
               // update hash code
               Xor(m_pawn_hash, start, blackPawn);
               //Xor(m_pawn_hash, target, whitePawn);
               Xor(m_pawn_hash, dest, blackPawn);
               target = old_epsq;
#ifdef _DEBUG
               targetPiece = my_Contents[target];
#endif
               my_Contents[dest] = blackPawn;
               pawn_bits[Black].set(dest);
               break;
            case Promotion:
               // update hash code
               Xor(m_pawn_hash, start, blackPawn);
               my_Contents[dest] = MakeBlackPiece(PromoteTo(move));
               my_Material[Black].remove_pawn();
               my_Material[Black].add_piece(PromoteTo(move));
               switch (PromoteTo(move))
               {
               case Knight:
                  knight_bits[Black].set(dest);
                  break;
               case Bishop:
                  bishop_bits[Black].set(dest);
                  bishopsqueens[Black].set(dest);
                  break;
               case Rook:
                  rook_bits[Black].set(dest);
                  rooksqueens[Black].set(dest);
                  break;
               case Queen:
                  queen_bits[Black].set(dest);
                  bishopsqueens[Black].set(dest);
                  rooksqueens[Black].set(dest);
                  break;
               default:
                  break;
               }
               break;
            default:
               Xor(m_pawn_hash, start, blackPawn);
               Xor(m_pawn_hash, dest, blackPawn);
               my_Contents[dest] = blackPawn;
               if (dest - start == 16) // 2-square pawn advance
               {
		 if (TEST_MASK(Bearing::ep_mask[File(dest)-1][(int)Black],pawn_bits[White])) {
                    my_EnPassantSq = dest;
		 }
               }
               pawn_bits[Black].set(dest);
               break;
            }
            pawn_bits[Black].clear(start);
            break;
         case Knight:
            my_Contents[dest] = blackKnight;
            knight_bits[Black].setClear(bits);
            break;
         case Bishop:
            my_Contents[dest] = blackBishop;
            bishop_bits[Black].setClear(bits);
            bishopsqueens[Black].setClear(bits);
            break;
         case Rook:
            my_Contents[dest] = blackRook;
            rook_bits[Black].setClear(bits);
            rooksqueens[Black].setClear(bits);
            UpdateCastleStatus(Black,start);
            break;
         case Queen:
            my_Contents[dest] = blackQueen;
            queen_bits[Black].setClear(bits);
            rooksqueens[Black].setClear(bits);
            bishopsqueens[Black].setClear(bits);
            break;
         case King:
            my_Contents[dest] = blackKing;
            my_KingPos[Black] = dest;
            if ((CastleStatus(side) != CastledQSide) &&
                (CastleStatus(side) != CastledKSide))
               my_CastleStatus[Black] = CantCastleEitherSide;
            break;
         }
         my_Contents[start] = EmptyPiece;
         if (Capture(move) != EmptyPiece)
         {
            move_count = 0;
            ASSERT(target != InvalidSquare);
#ifdef _DEBUG
            ASSERT(targetPiece != EmptyPiece &&
                   PieceColor(targetPiece) == White);
#endif
            occupied[White].clear(target);
            Piece cap = MakeWhitePiece(Capture(move));
            switch (Capture(move))
            {
            case Pawn:
               pawn_bits[White].clear(target);
               Xor(m_pawn_hash, target, cap);
               if (moveType == EnPassant)
               {
                  my_Contents[target] = EmptyPiece;
                  clearAll(White,target);
               }
               my_Material[White].remove_pawn();
               break;
            case Rook:
               UpdateCastleStatus(White,dest);
               rook_bits[White].clear(target);
               rooksqueens[White].clear(target);
               my_Material[White].remove_piece(Rook);
               break;
            case Knight:
               knight_bits[White].clear(target);
               my_Material[White].remove_piece(Knight);
               break;
            case Bishop:
               bishop_bits[White].clear(target);
               bishopsqueens[White].clear(target);
               my_Material[White].remove_piece(Bishop);
               break;
            case Queen:
               queen_bits[White].clear(target);
               rooksqueens[White].clear(target);
               bishopsqueens[White].clear(target);
               my_Material[White].remove_piece(Queen);
               break;
            case King:
               ASSERT(0);
               my_KingPos[White] = InvalidSquare;
               my_CastleStatus[White] = CantCastleEitherSide;
               my_Material[White].remove_piece(King);
               break;
            default: 
               break;
            }
         }
         setAll(Black,dest);
         clearAll(Black,start);
      }
   }

   ASSERT(rep_list_head-rep_list < (int)RepListSize);
   *rep_list_head++ = hashCode;
   ASSERT(pawn_hash() == BoardHash::pawnHash(*this));
   side = OppositeSide();
   all_occupied = occupied[White];
   all_occupied.Or(occupied[Black]);

#ifdef _DEBUG
   // verify correct updating of bitmaps:
   Board copy = *this;
   copy.set_secondary_vars();
   ASSERT(pawn_bits[White] == copy.pawn_bits[White]);
   ASSERT(knight_bits[White] == copy.knight_bits[White]);
   ASSERT(bishop_bits[White] == copy.bishop_bits[White]);
   ASSERT(rook_bits[White] == copy.rook_bits[White]);
   ASSERT(queen_bits[White] == copy.queen_bits[White]);
   ASSERT(bishopsqueens[White] == copy.bishopsqueens[White]);
   ASSERT(rooksqueens[White] == copy.rooksqueens[White]);
   ASSERT(occupied[White] == copy.occupied[White]);
   ASSERT(occupiedR90 == copy.occupiedR90);
   ASSERT(occupiedDa1 == copy.occupiedDa1);
   ASSERT(occupiedDa8 == copy.occupiedDa8);

   ASSERT(pawn_bits[Black] == copy.pawn_bits[Black]);
   ASSERT(knight_bits[Black] == copy.knight_bits[Black]);
   ASSERT(bishop_bits[Black] == copy.bishop_bits[Black]);
   ASSERT(rook_bits[Black] == copy.rook_bits[Black]);
   ASSERT(queen_bits[Black] == copy.queen_bits[Black]);
   ASSERT(bishopsqueens[Black] == copy.bishopsqueens[Black]);
   ASSERT(rooksqueens[Black] == copy.rooksqueens[Black]);
   ASSERT(occupied[Black] == copy.occupied[Black]);
   ASSERT(occupiedR90 == copy.occupiedR90);
   ASSERT(my_Contents[my_KingPos[White]]==whiteKing);
   ASSERT(my_Contents[my_KingPos[Black]]==blackKing);
#endif     
}

void Board::undo_castling(const Square &kp,
                          const Square &oldkingsq, const Square &newrooksq,
                          const Square &oldrooksq)
{
   my_Contents[kp] = EmptyPiece;
   my_Contents[oldrooksq] = MakePiece(Rook,side);
   my_Contents[newrooksq] = EmptyPiece;
   my_Contents[oldkingsq] = MakePiece(King,side);
   my_KingPos[side] = oldkingsq;
   rook_bits[side].set(oldrooksq);
   rook_bits[side].clear(newrooksq);
   rooksqueens[side].set(oldrooksq);
   rooksqueens[side].clear(newrooksq);

   setAll(side,oldrooksq);
   setAll(side,oldkingsq);
   clearAll(side,kp);
   clearAll(side,newrooksq);
}

void Board::UndoMove( const Move rmove, const Board_State &state )
{
   side = OppositeColor(side);
   if (!IsNull(rmove))
   {
      const SpecialType moveType = TypeOfMove(rmove);
      const Square start = StartSquare(rmove);
      const Square dest = DestSquare(rmove);
      if (moveType == KCastle)
      {
         Square kp = KingPos(side);
         Square oldrooksq = kp+1;
         Square newrooksq = kp-1;
         Square oldkingsq = kp-2;
         undo_castling(kp,oldkingsq,newrooksq,oldrooksq);
      }
      else if (moveType == QCastle)
      {
         Square kp = KingPos(side);
         Square oldrooksq = kp-2;
         Square newrooksq = kp+1;
         Square oldkingsq = kp+2;
         undo_castling(kp,oldkingsq,newrooksq,oldrooksq);
      }
      else if (side == White)
      {
         const Bitmap bits(Bitmap::mask[start] |
                           Bitmap::mask[dest]);
         // not castling
         Square target;
         if (moveType == EnPassant)
            target = DestSquare(rmove) + RankIncr;
         else
            target = DestSquare(rmove);
         // fix up start square:
         if (moveType == Promotion)
         {
            my_Contents[start] = whitePawn;
         }
         else
         {
            my_Contents[start] = MakeWhitePiece(PieceMoved(rmove));
         }
         setAll(White,start);
         switch (PieceMoved(rmove))
         {
         case Pawn:
            if (moveType == Promotion)
            {
               my_Material[White].add_pawn();
               my_Material[White].remove_piece(PromoteTo(rmove));
               switch (PromoteTo(rmove))
               {
               case Knight:
                  knight_bits[White].clear(dest);
                  break;
               case Bishop:
                  bishop_bits[White].clear(dest);
                  bishopsqueens[White].clear(dest);
                  break;
               case Rook:
                  rook_bits[White].clear(dest);
                  rooksqueens[White].clear(dest);
                  break;
               case Queen:
                  queen_bits[White].clear(dest);
                  bishopsqueens[White].clear(dest);
                  rooksqueens[White].clear(dest);
                  break;
               default:
                  break;
               }
            }
            else
            {
               pawn_bits[White].clear(dest);
            }
            pawn_bits[White].set(start);
            break;
         case Knight:
            knight_bits[White].setClear(bits);
            break;
         case Bishop:
            bishop_bits[White].setClear(bits);
            bishopsqueens[White].setClear(bits);
            break;
         case Rook:
            rook_bits[White].setClear(bits);
            rooksqueens[White].setClear(bits);
            break;
         case Queen:
            queen_bits[White].setClear(bits);
            rooksqueens[White].setClear(bits);
            bishopsqueens[White].setClear(bits);
            break;
         case King:
            my_KingPos[White] = start;
            break;
         default:
            break;
         }
         // fix up dest square
         clearAll(White,dest);
         my_Contents[dest] = EmptyPiece;
         my_Contents[target] = MakePiece(Capture(rmove),Black);
         my_EnPassantSq = InvalidSquare;
         if (Capture(rmove) != EmptyPiece)
         {
            switch (Capture(rmove))
            {
            case Pawn:
               pawn_bits[Black].set(target);
               my_Material[Black].add_pawn();
               break;
            case Knight:
               knight_bits[Black].set(target);
               my_Material[Black].add_piece(Knight);
               break;
            case Bishop:
               bishop_bits[Black].set(target);
               bishopsqueens[Black].set(target);
               my_Material[Black].add_piece(Bishop);
               break;
            case Rook:
               rook_bits[Black].set(target);
               rooksqueens[Black].set(target);
               my_Material[Black].add_piece(Rook);
               break;
            case Queen:
               queen_bits[Black].set(target);
               bishopsqueens[Black].set(target);
               rooksqueens[Black].set(target);
               my_Material[Black].add_piece(Queen);
               break;
            case King:
               my_KingPos[Black] = target;
               my_Material[Black].add_piece(King);
               break;
            default:
               break;
            }
            setAll(Black,target);
         }
      }
      else // side == Black
      {
         const Bitmap bits(Bitmap::mask[start] |
                           Bitmap::mask[dest]);
         // not castling
         Square target;
         if (moveType == EnPassant)
            target = dest - RankIncr;
         else
            target = dest;
         // fix up start square:
         if (moveType == Promotion)
         {
            my_Contents[start] = blackPawn;
         }
         else
         {
            my_Contents[start] = MakeBlackPiece(PieceMoved(rmove));
         }
         setAll(Black,start);
         switch (PieceMoved(rmove))
         {
         case Pawn:
            if (moveType == Promotion)
            {
               my_Material[Black].add_pawn();
               my_Material[Black].remove_piece(PromoteTo(rmove));
               switch (PromoteTo(rmove))
               {
               case Knight:
                  knight_bits[Black].clear(dest);
                  break;
               case Bishop:
                  bishop_bits[Black].clear(dest);
                  bishopsqueens[Black].clear(dest);
                  break;
               case Rook:
                  rook_bits[Black].clear(dest);
                  rooksqueens[Black].clear(dest);
                  break;
               case Queen:
                  queen_bits[Black].clear(dest);
                  bishopsqueens[Black].clear(dest);
                  rooksqueens[Black].clear(dest);
                  break;
               default:
                  break;
               }
            }
            else
            {
               pawn_bits[Black].clear(dest);
            }
            pawn_bits[Black].set(start);
            break;
         case Knight:
            knight_bits[Black].setClear(bits);
            break;
         case Bishop:
            bishop_bits[Black].setClear(bits);
            bishopsqueens[Black].setClear(bits);
            break;
         case Rook:
            rook_bits[Black].setClear(bits);
            rooksqueens[Black].setClear(bits);
            break;
         case Queen:
            queen_bits[Black].setClear(bits);
            rooksqueens[Black].setClear(bits);
            bishopsqueens[Black].setClear(bits);
            break;
         case King:
            my_KingPos[Black] = start;
            break;
         default:
            break;
         }
         // fix up dest square
         clearAll(Black,dest);
         my_Contents[dest] = EmptyPiece;
         my_Contents[target] = MakePiece(Capture(rmove),White);
         my_EnPassantSq = InvalidSquare;
         if (Capture(rmove) != EmptyPiece)
         {
            switch (Capture(rmove))
            {
            case Pawn:
               pawn_bits[White].set(target);
               my_Material[White].add_pawn();
               break;
            case Knight:
               knight_bits[White].set(target);
               my_Material[White].add_piece(Knight);
               break;
            case Bishop:
               bishop_bits[White].set(target);
               bishopsqueens[White].set(target);
               my_Material[White].add_piece(Bishop);
               break;
            case Rook:
               rook_bits[White].set(target);
               rooksqueens[White].set(target);
               my_Material[White].add_piece(Rook);
               break;
            case Queen:
               queen_bits[White].set(target);
               bishopsqueens[White].set(target);
               rooksqueens[White].set(target);
               my_Material[White].add_piece(Queen);
               break;
            case King:
               my_KingPos[White] = target;
               my_Material[White].add_piece(King);
               break;
            default:
               break;
            }
            setAll(White,target);
         }
      }
   }
   state.restoreState(*this);
   
   --rep_list_head;
   all_occupied = occupied[White];
   all_occupied.Or(occupied[Black]);
#ifdef _DEBUG
   // verify correct updating of bitmaps:
   Board copy = *this;
   copy.set_secondary_vars();
   ASSERT(pawn_bits[White] == copy.pawn_bits[White]);
   ASSERT(knight_bits[White] == copy.knight_bits[White]);
   ASSERT(bishop_bits[White] == copy.bishop_bits[White]);
   ASSERT(rook_bits[White] == copy.rook_bits[White]);
   ASSERT(queen_bits[White] == copy.queen_bits[White]);
   ASSERT(bishopsqueens[White] == copy.bishopsqueens[White]);
   ASSERT(rooksqueens[White] == copy.rooksqueens[White]);
   ASSERT(occupied[White] == copy.occupied[White]);
   ASSERT(occupiedR90 == copy.occupiedR90);
   ASSERT(occupiedDa1 == copy.occupiedDa1);
   ASSERT(occupiedDa8 == copy.occupiedDa8);

   ASSERT(pawn_bits[Black] == copy.pawn_bits[Black]);
   ASSERT(knight_bits[Black] == copy.knight_bits[Black]);
   ASSERT(bishop_bits[Black] == copy.bishop_bits[Black]);
   ASSERT(rook_bits[Black] == copy.rook_bits[Black]);
   ASSERT(queen_bits[Black] == copy.queen_bits[Black]);
   ASSERT(bishopsqueens[Black] == copy.bishopsqueens[Black]);
   ASSERT(rooksqueens[Black] == copy.rooksqueens[Black]);
   ASSERT(occupied[Black] == copy.occupied[Black]);
   ASSERT(occupiedR90 == copy.occupiedR90);
   ASSERT(my_Contents[my_KingPos[White]]==whiteKing);
   ASSERT(my_Contents[my_KingPos[Black]]==blackKing);

#endif     
}

int Board::any_attacks(const Square sq, const ColorType side) const
{
   if (sq == InvalidSquare)
      return 0;

   if (side == White)
   {
      if (TEST_MASK(Bearing::pawn_attacks[sq][White],pawn_bits[White]))
         return 1;

      if (TEST_MASK(Bearing::knight_attacks[sq],knight_bits[White]))
         return 1;

      if (Bearing::king_attacks[sq].is_set(KingPos(White)))
         return 1;
   }
   else
   {
      if (TEST_MASK(Bearing::pawn_attacks[sq][Black],pawn_bits[Black]))
         return 1;

      if (TEST_MASK(Bearing::knight_attacks[sq],knight_bits[Black]))
         return 1;

      if (Bearing::king_attacks[sq].is_set(KingPos(Black)))
         return 1;
   }

   // Find attacks on the rank
   const int r = ((int)sq)/8;
   const int f = File(sq)-1;
   if
       (TEST_MASK(rooksqueens[side],Bearing::rank_mask[r]))
   {
      // See if any rooks or queens are there
      if (TEST_MASK(RankAttacks(*this,sq),rooksqueens[side]))
         return 1;
   }

   if (TEST_MASK(rooksqueens[side],Bearing::file_mask[f]))
   {
      if (TEST_MASK(FileAttacks(*this,sq),rooksqueens[side]))
         return 1;
   }

   if (TEST_MASK(bishopsqueens[side],Bearing::diag_a1_mask[sq]))
   {
      // Now find attacks on the a1-h8 diagonal
      if (TEST_MASK(DiagAtcksA1(*this,sq),bishopsqueens[side]))
         return 1;
   }

   // Now find attacks on the a8-h1 diagonal
   // See if any bishops or queens are there
   if (TEST_MASK(bishopsqueens[side],Bearing::diag_a8_mask[sq]))
   {
      if (TEST_MASK(DiagAtcksA8(*this,sq),bishopsqueens[side]))
         return 1;
   }
   return 0;
}

int Board::any_attacks(const Square sq, const ColorType side,
                        Bitmap &source) const
{
   if (sq == InvalidSquare)
      return 0;

   source = Bearing::pawn_attacks[sq][side];
   source.And(pawn_bits[side]);
   if (!source.is_clear())
      return 1;

   source = Bearing::knight_attacks[sq];
   source.And(knight_bits[side]);
   if (!source.is_clear())
      return 1;

   source = Bearing::king_attacks[sq];
   if (source.is_set(KingPos(White)))
      return 1;

   // Find attacks on the rank
   const int r = ((int)sq)/8;
   const int f = File(sq)-1;
   if
       (TEST_MASK(rooksqueens[side],Bearing::rank_mask[r]))
   {
      source = Bitmap::And(RankAttacks(*this,sq),rooksqueens[side]);
      if (!source.is_clear())
         return 1;
   }

   if (TEST_MASK(rooksqueens[side],Bearing::file_mask[f]))
   {
      if (!Bitmap::And(FileAttacks(*this,sq),rooksqueens[side]).is_clear())
         return 1;
   }

   // Now find attacks on the a1-h8 diagonal
   // find the pieces nearest to sq on the diagonal, then unrotate the bits
   if (TEST_MASK(bishopsqueens[side],Bearing::diag_a1_mask[sq]))
   {
      source = DiagAtcksA1(*this,sq);
      // See if any bishops or queens are there
      if (!source.is_clear())
         return 1;
   }

   if (TEST_MASK(bishopsqueens[side],Bearing::diag_a8_mask[sq]))
   {
      // Now find attacks on the a8-h1 diagonal
      source = DiagAtcksA8(*this,sq);
      if (!source.is_clear())
         return 1;
   }
   return 0;
}

Bitmap Board::calc_attacks(const Square sq, const ColorType side) const
{
   Bitmap retval;

   Bitmap patcks(Bearing::pawn_attacks[sq][side]);
   patcks.And(pawn_bits[side]);
   retval.Or(patcks);

   Bitmap natcks(Bearing::knight_attacks[sq]);
   natcks.And(knight_bits[side]);
   retval.Or(natcks);

   Bitmap katcks(Bearing::king_attacks[sq]);
   Bitmap mask;
   mask.set(KingPos(side));
   katcks.And(mask);
   retval.Or(katcks);

   const int r = ((int)sq)/8;
   const int f = File(sq)-1;
   if
       (TEST_MASK(rooksqueens[side],Bearing::rank_mask[r]))
   {
      retval.data |= Bitmap::And(RankAttacks(*this,sq),rooksqueens[side]).data;
   }

   if (TEST_MASK(rooksqueens[side],Bearing::file_mask[f]))
   {
      // To find attacks on the file, use the 90-degree rotated bitmaps.
      // Same procedure as for ranks, above.
      // shift the occupied bits so that the file that sq occupies are
      // in bits 0-7.
      // find the pieces nearest to sq on the file, then
      // "unrotate" the result and shift it back up.
      // See if any rooks or queens are there
      retval.data |= Bitmap::And(FileAttacks(*this,sq),rooksqueens[side]).data;
   }

   // Now find attacks on the a1-h8 diagonal
   // find the pieces nearest to sq on the diagonal, then unrotate the bits
   if (TEST_MASK(bishopsqueens[side],Bearing::diag_a1_mask[sq]))
   {
      retval.data |= Bitmap::And(DiagAtcksA1(*this,sq),
                            bishopsqueens[side]).data;
   }
   if (TEST_MASK(bishopsqueens[side],Bearing::diag_a8_mask[sq]))
   {
      // Now find attacks on the a8-h1 diagonal
      retval.data |= Bitmap::And(DiagAtcksA8(*this,sq),
                            bishopsqueens[side]).data;
   }
   return retval;
}

Bitmap Board::min_attacker(const Bitmap &atcks, const ColorType side)
const
{
   Bitmap retval;
   if (side == White)
   {
      retval = Bitmap::And(atcks,pawn_bits[White]);
      if (!retval.is_clear())
         return retval;
      retval = Bitmap::And(atcks,knight_bits[White]);
      if (!retval.is_clear())
         return retval;
      retval = Bitmap::And(atcks,bishop_bits[White]);
      if (!retval.is_clear())
         return retval;
      retval = Bitmap::And(atcks,rook_bits[White]);
      if (!retval.is_clear())
         return retval;
      retval = Bitmap::And(atcks,queen_bits[White]);
      if (!retval.is_clear())
         return retval;
      if (atcks.is_set(KingPos(White)))
	return atcks;
      else
        return Bitmap(0);
   }
   else
   {
      retval = Bitmap::And(atcks,pawn_bits[Black]);
      if (!retval.is_clear())
         return retval;
      retval = Bitmap::And(atcks,knight_bits[Black]);
      if (!retval.is_clear())
         return retval;
      retval = Bitmap::And(atcks,bishop_bits[Black]);
      if (!retval.is_clear())
         return retval;
      retval = Bitmap::And(atcks,rook_bits[Black]);
      if (!retval.is_clear())
         return retval;
      retval = Bitmap::And(atcks,queen_bits[Black]);
      if (!retval.is_clear())
         return retval;
      if (atcks.is_set(KingPos(Black)))
	return atcks;
      else
        return Bitmap(0);
   }    
   return retval;
}

Bitmap Board::all_pawn_attacks( const ColorType side) const
{
   if (side == White)
   {
      Bitmap pawns1(pawn_bits[White]);
      Bitmap pawns2(pawns1);
      pawns1.shr(7);
      pawns1.andNot(Bitmap(MAKELONGLONG(0x01010101,0x01010101)));
      pawns2.shr(9);
      pawns2.andNot(Bitmap(MAKELONGLONG(0x80808080,0x80808080)));
      return Bitmap::Or(pawns1,pawns2);
   }
   else
   {
      Bitmap pawns1(pawn_bits[Black]);
      Bitmap pawns2(pawns1);
      pawns1.shl(7);
      pawns1.andNot(Bitmap(MAKELONGLONG(0x80808080,0x80808080)));
      pawns2.shl(9);
      pawns2.andNot(Bitmap(MAKELONGLONG(0x01010101,0x01010101)));
      return Bitmap::Or(pawns1,pawns2);
   }
}

int Board::rook_mobility(const Square &sq) const
{
   // We consider only horizontal mobility (for now)
   int mobl;
   Bitmap occ(all_occupied);
   // Shift so that the relevant bits are in bits 0-7 of the bitmap.
   int r = ((int)sq)/8;
   mobl = Bearing::mobility[
       Bearing::whichAttack[(occ.rank_bits(r)) & 0xff][File(sq)-1]];
   return mobl;
}

Board::CheckStatusType Board::CheckStatus() const
{
   if (my_CheckStatus != Unknown)
   {
      return my_CheckStatus;
   }
   else if (any_attacks(KingPos(Side()),OppositeSide()))
   {
      // This is a const function, but we cache its result
      Board &b = (Board&)*this;
      b.my_CheckStatus = InCheck;
   }
   else
   {
      // This is a const function, but we cache its result
      Board &b = (Board&)*this;
      b.my_CheckStatus = NotInCheck;
   }
   return my_CheckStatus;
}

// This variant of CheckStatus sees if the last move made
// delivered check. It is generally faster than CheckStatus
// with no param, because we can use the last move information
// to avoid calling any_attacks, in many cases.
Board::CheckStatusType Board::CheckStatus(Move lastMove) const
{
   if (my_CheckStatus != Unknown)
   {
      return my_CheckStatus;
   }
   if (IsNull(lastMove))
      return CheckStatus();
   Square kp = my_KingPos[Side()];
   Square checker = DestSquare(lastMove);
   int d = Bearing::Directions[checker][kp];
   switch(PieceMoved(lastMove)) {
      case Pawn: {
          if (TypeOfMove(lastMove) != Normal)
             return CheckStatus();
          if (Bearing::pawn_attacks[kp][OppositeSide()].is_set(checker)) {
              Board &b = (Board&)*this;
              b.my_CheckStatus = InCheck;
          }
          else if (Bearing::Directions[kp][StartSquare(lastMove)] == 0) {
              Board &b = (Board&)*this;
              b.my_CheckStatus = NotInCheck;
          }
          if (my_CheckStatus == Unknown)
              return CheckStatus();
          else
              return my_CheckStatus;
       }
       case Rook:
        switch(d) {
          case 1:
              if (RankAttacksRight(*this,checker).is_set(kp)) {
                 Board &b = (Board&)*this;
                 b.my_CheckStatus = InCheck;
              }
              break;
          case -1:
              if (RankAttacksLeft(*this,checker).is_set(kp)) {
                 Board &b = (Board&)*this;
                 b.my_CheckStatus = InCheck;
              }
              break;
          case 8:
              if (FileAttacksDown(*this,checker).is_set(kp)) {
                 Board &b = (Board&)*this;
                 b.my_CheckStatus = InCheck;
              }
              break;
          case -8:
              if (FileAttacksUp(*this,checker).is_set(kp)) {
                 Board &b = (Board&)*this;
                 b.my_CheckStatus = InCheck;
              }
          default:
            break;
        }
        if (my_CheckStatus == Unknown) {
          // see if move could generate a discovered check
            d = Util::Abs(Bearing::Directions[kp][StartSquare(lastMove)]);
            if (d == 0 || d == 8 || d == 1) {
                 Board &b = (Board&)*this;
                 b.my_CheckStatus = NotInCheck;
                 return NotInCheck;
            }
            else
                return CheckStatus();
        }
        else
            return my_CheckStatus;
        break;
      case Bishop:
        switch(d) {
          case 7:
              if (DiagAtcksA1Lower(*this,checker).is_set(kp)) {
                 Board &b = (Board&)*this;
                 b.my_CheckStatus = InCheck;
              }
              break;
          case -7:
              if (DiagAtcksA1Upper(*this,checker).is_set(kp)) {
                 Board &b = (Board&)*this;
                 b.my_CheckStatus = InCheck;
              }
              break;
          case 9:
              if (DiagAtcksA8Lower(*this,checker).is_set(kp)) {
                 Board &b = (Board&)*this;
                 b.my_CheckStatus = InCheck;
              }
              break;
          case -9:
              if (DiagAtcksA8Upper(*this,checker).is_set(kp)) {
                 Board &b = (Board&)*this;
                 b.my_CheckStatus = InCheck;
              }
          default:
            break;
        }
        if (my_CheckStatus == Unknown) {
          // see if move could generate a discovered check
            d = Util::Abs(Bearing::Directions[kp][StartSquare(lastMove)]);
            if (d == 0 || d == 7 || d == 9) {
                 Board &b = (Board&)*this;
                 b.my_CheckStatus = NotInCheck;
                 return NotInCheck;
            }
            else
                return CheckStatus();
        }
        else
            return my_CheckStatus;
        break;
      case Knight:
        if (Bearing::knight_attacks[checker].is_set(kp))
        {
           Board &b = (Board&)*this;
           b.my_CheckStatus = InCheck;
        }
        else {
            d = Util::Abs(Bearing::Directions[kp][StartSquare(lastMove)]);
            if (d == 0)
            {
               Board &b = (Board&)*this;
               b.my_CheckStatus = NotInCheck;
            }
            else
               return CheckStatus();
        }
        break;
      case King: 
          if (TypeOfMove(lastMove) != Normal) /* castling */
              return CheckStatus();
          if (Bearing::king_attacks[checker].is_set(kp)) {
            Board &b = (Board&)*this;
            b.my_CheckStatus = InCheck;
          }
          else if (Bearing::Directions[StartSquare(lastMove)][kp] == 0) {
            Board &b = (Board&)*this;
            b.my_CheckStatus = NotInCheck;
          }
          else
            return CheckStatus();
      default:
          break;
   }
   return CheckStatus();
}
   
static void set_bad( istream &i )
{
   i.clear( ios::badbit | i.rdstate() );
}

int Board::rep_count() const
{
   int entries = move_count-2;
   hash_t to_match = HashCode();
   int count = 0;
   if (entries == 0) return 0;
   for (hash_t *rep_list=rep_list_head-3;
       entries>=0;
       rep_list-=2,entries-=2)
   {
      if (*rep_list == to_match)
      {
         count++;
         if (count >= 2)
         {
            return count;
         }
      }
   }
   return count;
}

istream & operator >> (istream &i, Board &board)
{
   // read in a board position in Forsythe-Edwards (FEN) notation.
   static char buf[128];

   char *bp = buf;
   int c;
   int fields = 0; int count = 0;
   while (i.good() && fields < 4 && (c = i.get()) != '\n' && 
          c != -1 && 
          ++count < 128)
   {
      *bp++ = c;
      if (isspace(c))
         fields++;
   }
   *bp = '\0';
   if (!i)
      return i;
   if (!BoardIO::read_fen(board, buf))
      set_bad(i);
   return i;
}

ostream & operator << (ostream &o, const Board &board)
{
   char buf[128];
   memset(buf,'\0',128);
   BoardIO::write_fen(board,buf,1);
   o << buf;
   return o;
}

