// Copyright 1994-1998 by Jon Dart.  All Rights Reserved.

#include "moveord.h"
#include "movegen.h"
#include "scoring.h"
#include "constant.h"
#include "attacke.h"
#include "bearing.h"
#include "util.h"
#include "globals.h"
#include "config.h"
#include "boardsta.h"

#include <iostream>

static Move Killers1[Constants::MaxPly];
static Move Killers2[Constants::MaxPly];
static int HistoryW[64][64];
static int HistoryB[64][64];

inline void swap( Move moves[], const int i, const int j)
{
   Move tmp = moves[j];
   moves[j] = moves[i]; 
   moves[i] = tmp;
}

// move scoring routine, used only at ply 0
int Move_Ordering::score(Board &board, const Move &move )
{
   int val = 0;
   Square dest_square = DestSquare(move);
   Piece capture = board[dest_square];
   if (!IsEmptyPiece(capture))
   {
      if (!board.any_attacks(dest_square, board.OppositeSide()))
         // attack on undefended piece
      {
         val += PieceValue(capture);
      }
      else if (TypeOfMove(move) == Normal)
      {
         int est = attack_estimate(board,move);
         if (est > 0)
            val += est;
         else // losing capture
            val += 10 + (est/16);
      }
      else if (TypeOfMove(move) == Promotion)
      {
         val += PieceValue(PromoteTo(move)) - PieceValue(Pawn);
      }
      else // en-passant?
         val += 32;
   }
   if (EqualMove(move,Killers1[0]) || EqualMove(move,Killers2[0]))
      val += 30;
   Board_State stat(board);
   board.DoMove(move);
   if (board.any_attacks(board.KingPos(board.OppositeSide()),board.Side()))
   {
      // move into check, illegal
      val = -Constants::BIG;
   }
   else
   {
      val -= Scoring::positional_score(board,-Constants::BIG,Constants::BIG);
      if (board.CheckStatus() == Board::InCheck)
         val += 25;
   }
   board.UndoMove(move,stat);
#ifdef _TRACE
   cout << MoveImage(move) << " score = " << val << endl;
#endif
   return val;
}

int
Move_Ordering::order_moves(const Board & board,
                           Move moves[], const int num_moves,
                           const int ply, const Move & best_move,
                           int &top)
{
   top = 0;
   if (num_moves == 0)
      return 0;
   // Move gainful capture moves and killers to the front
   // of the move list.
   static int scores[Move_Generator::MaxMoves];
   int j = 0, i = 0;
#ifdef _TRACE
   int k;
#endif
   for (i = 0; i < num_moves; i++)
   {
      int score = 0;
      const Square dest = DestSquare(moves[i]);
      const Square start = StartSquare(moves[i]);
      const Piece capture = board[dest];

      if (EqualMove(moves[i],best_move))
      {
         scores[j] = Constants::BIG;
#ifdef _TRACE
         for (k=0; k < ply; k++) cout << ' ';
         cout << "move " << MoveImage(moves[i]) << " (pv)"<< endl;
#endif
         swap(moves,i,j);
         ++j;
      }
      else if (!IsEmptyPiece(capture))
      {
         int start_val;
         //if (TypeOfPiece(board[start]) == King)
         //    start_val = PieceValue(Pawn);
         //else
         start_val = PieceValue(board[start]);
         int capture_val = PieceValue(capture);
         int gain;
         if (capture_val > start_val && PromoteTo(moves[i]) == InvalidPiece)
            gain = capture_val - start_val;
         else
            gain = attack_estimate(board,moves[i]);
#ifdef _TRACE
         for (k=0; k < ply; k++) cout << ' ';
         cout << "move " << MoveImage(moves[i]) << " est = " << gain << endl;
#endif
         if (gain >= -10)
         {
            if (i != j)
            {
               swap(moves,i,j);
            }
            scores[j] = gain;
            ++j;
         }
      }
      else if (TypeOfMove(moves[i]) == Promotion)
      {
         score = PieceValue(PromoteTo(moves[i])) - 64;
#ifdef _TRACE
         for (k=0; k < ply; k++) cout << ' ';
         cout << "move " << MoveImage(moves[i]) << " est = " << score << endl;
#endif
         if (score > 0)
         {
            scores[j] = score;
            swap(moves,i,j);
            ++j;
         }
      }
      else
      {
         // not a capture move
         if ((StartSquare(moves[i]) == StartSquare(Killers1[ply]) &&
              DestSquare(moves[i]) == DestSquare(Killers1[ply])) ||
             (StartSquare(moves[i]) == StartSquare(Killers2[ply]) &&
              DestSquare(moves[i]) == DestSquare(Killers2[ply])))
         //            if (EqualMove(moves[i],Killers1[ply]) ||
         //                EqualMove(moves[i],Killers2[ply]))
         {
#ifdef _TRACE
            for (k=0; k < ply; k++) cout << ' ';
            cout << "move " << MoveImage(moves[i]) << " (killer)" << endl;
#endif
            scores[j] = -40; // put killers after captures
            swap(moves,i,j);
            ++j;
         }
      }
   }
   if (j > 1)
      sort_moves(moves,scores,j);
#ifdef _TRACE
   for (k=0; k < ply ; k++) cout << ' ';
   cout << "after sort 1: " ;
   for (k=0; k <j; k++)
      cout << MoveImage(moves[k]) << ' ';
   cout << endl;
#endif
   top = j;
   return num_moves;
}

int
Move_Ordering::reorder_ply0_moves(const Board & board,
                           Move moves[], const int num_moves,
                           const Move & best_move)
{
   if (num_moves == 0)
      return 0;
   // Move gainful capture moves and killers to the front
   // of the move list.
   static int scores[Move_Generator::MaxMoves];
   int j = 0, i = 0;
#ifdef _TRACE
   int k;
#endif
   for (i = 0; i < num_moves; i++)
   {
      int score = 0;
      const Square dest = DestSquare(moves[i]);
      const Square start = StartSquare(moves[i]);
      const Piece capture = board[dest];

      if (EqualMove(moves[i],best_move))
      {
         scores[j] = Constants::BIG;
#ifdef _TRACE
         cout << "move " << MoveImage(moves[i]) << " (pv)"<< endl;
#endif
         swap(moves,i,j);
         ++j;
      }
      else if (!IsEmptyPiece(capture))
      {
         int start_val;
         //if (TypeOfPiece(board[start]) == King)
         //    start_val = PieceValue(Pawn);
         //else
         start_val = PieceValue(board[start]);
         int capture_val = PieceValue(capture);
         int gain;
         if (capture_val > start_val && PromoteTo(moves[i]) == InvalidPiece)
            gain = capture_val - start_val;
         else
            gain = attack_estimate(board,moves[i]);
#ifdef _TRACE
         cout << "move " << MoveImage(moves[i]) << " est = " << gain << endl;
#endif
         if (gain >= -10)
         {
            if (i != j)
            {
               swap(moves,i,j);
            }
            scores[j] = gain;
            ++j;
         }
      }
      else if (TypeOfMove(moves[i]) == Promotion)
      {
         score = PieceValue(PromoteTo(moves[i])) - 64;
#ifdef _TRACE
         cout << "move " << MoveImage(moves[i]) << " est = " << score << endl;
#endif
         if (score > 0)
         {
            scores[j] = score;
            swap(moves,i,j);
            ++j;
         }
      }
      else
      {
         // not a capture move
         if ((StartSquare(moves[i]) == StartSquare(Killers1[0]) &&
              DestSquare(moves[i]) == DestSquare(Killers1[0]))) 
         {
#ifdef _TRACE
            cout << "move " << MoveImage(moves[i]) << " (killer)" << endl;
#endif
            scores[j] = 5000; // put killers before captures
            swap(moves,i,j);
            ++j;
         }
         else if ((StartSquare(moves[i]) == StartSquare(Killers2[0]) &&
              DestSquare(moves[i]) == DestSquare(Killers2[0]))) {
#ifdef _TRACE
            cout << "move " << MoveImage(moves[i]) << " (killer)" << endl;
#endif
            scores[j] = 4999; // put killers before captures
            swap(moves,i,j);
            ++j;

         }
      }
   }
   if (j > 1)
      sort_moves(moves,scores,j);
#ifdef _TRACE
   cout << "after ply 0 resort: " ;
   for (k=0; k <j; k++)
      cout << MoveImage(moves[k]) << ' ';
   cout << endl;
#endif
   return num_moves;
}

void Move_Ordering::order_moves2(const Board &board,Move moves[], int top, int num_moves)
{
  static int scores[Move_Generator::MaxMoves];
  if (options.search.use_history && num_moves > top)
    {
      int max = 0, maxi=-1;
      int maxi2 =-1;
      for (int i = top; i < num_moves; i++)
	{
	  if (board.Side() == White)
	    {
	      scores[i] = HistoryW[StartSquare(moves[i])][DestSquare(moves[i])];
	    }
         else
	   {
	     scores[i] = HistoryB[StartSquare(moves[i])][DestSquare(moves[i])];
	   }
	  if (scores[i] > max)
	    {
	      if (maxi >=0) {
		maxi2 = maxi;
	      }
	      maxi = i;
	      max = scores[i];
	    }
	}
      //      sort_moves(moves+top,scores+top,num_moves-top);
      if (maxi != -1) swap(moves,top,maxi);
      if (maxi2 != -1) swap(moves,top+1,maxi2);
    }
}


int Move_Ordering::order_ply0_moves( Board &board, Move moves[],
                                     const int num_moves, const Move &best_move, int &easy)
{
   Scoring::init(board);
   int scores[Move_Generator::MaxMoves];

   int i;
   for (i = 0; i < num_moves; i++)
   {
      if (EqualMove(moves[i],best_move))
         scores[i] = Constants::BIG;
      else
         scores[i] = score(board,moves[i]);
   }
   sort_moves(moves, scores, num_moves);
   int real_num_moves = num_moves;

   easy = 1;
   for (i=num_moves-1; i >=0; --i)
   {
      if (scores[i] == -Constants::BIG)
      {
         --real_num_moves;
      }
      else if (i >0 && scores[0]-100 < scores[i])
	easy = 0;
   }
   return real_num_moves;
}

int
Move_Ordering::order_captures(Board & board,
                              Move moves[], const int num_moves,
                              const int alpha, const int material)
{
   static int scores[Move_Generator::MaxMoves];
   int n=0;
   int bias = QSEARCH_FORWARD_PRUNE_BIAS;
  
   for (int i = 0; i < num_moves; i++)
   {
      PieceType capture = Capture(moves[i]);
      Square loc = StartSquare(moves[i]);
      if (capture != EmptyPiece)
      {
         int est;
         int start_val = (int)PieceValue(board[loc]);
         if (capture == King) // king capture
         {
            // There is no point sorting the
            // rest of the moves. We return only
            // this move, which signals that
            // the previous move was illegal.
            moves[0] = moves[i];
            return 1;
         }
         else
         {
            est = (int)PieceValue(capture);
         }
         if (PromoteTo(moves[i]) != InvalidPiece)
            est += (int)PieceValue(PromoteTo(moves[i])) - PAWN_VALUE;
         // Futility check: if capturing the piece "for free" will not
         // get our score up to alpha, discard the move
         if (material + bias + est >= alpha)
         {
            est -= start_val;
            if (est > 10) 
            {
               // Always consider capturing a higher-valued piece
               scores[i] = est + (int)capture;
               if (i>n){ 
                 scores[n] = scores[i];
                 moves[n] = moves[i];
               }
               ++n;
            }
            else
            {
               // This capture appears to lose, but we call attack_estimate
               // to make sure
	           est = attack_estimate(board,moves[i]);
               if (bias + material + est >= alpha) {
                  scores[i] = est + (int)capture;
                  if (i>n){ 
                    scores[n] = scores[i];
                    moves[n] = moves[i];
                  }
                  ++n;
               }
            }
         }
      }
      else if (TypeOfMove(moves[i]) == Promotion)
      {
         scores[i] = PieceValue(PromoteTo(moves[i])) - 64;
         if (scores[i] + 64 + QSEARCH_FORWARD_PRUNE_BIAS + 
             material >= alpha) {
            if (i>n){ 
               scores[n] = scores[i];
               moves[n] = moves[i];
            }
            ++n;
         }
      }
   }
   if (n>1) Move_Ordering::sort_moves(moves, scores, n);
   return n;
}

void 
Move_Ordering::clear_killer()
{
   for (int i = 0; i < Constants::MaxPly; i++)
   {
      Killers1[i] =
      Killers2[i] = NullMove;   
   }
}

void 
Move_Ordering::clear_history()
{
   memset(HistoryW,'\0',4096*sizeof(int));
   memset(HistoryB,'\0',4096*sizeof(int));
}

void 
Move_Ordering::set_killer(const Board& board, const Move & move, 
                          const unsigned ply, const unsigned depth)
{
   if (options.search.num_killers > 1)
   {
      Killers2[ply] = Killers1[ply];
   }
   if (options.search.num_killers > 0)
   {
      Killers1[ply] = move;
   }
}

void 
Move_Ordering::update_history(const Board& board, const Move & move, 
                              const unsigned depth)
{
   if (options.search.use_history)
   {
      if (Capture(move) != EmptyPiece || PromoteTo(move) != InvalidPiece) return;
      if (board.Side() == White)
         HistoryW[(int)StartSquare(move)][(int)DestSquare(move)] += depth;
      else
         HistoryB[(int)StartSquare(move)][(int)DestSquare(move)] += depth;
   }
}

int Move_Ordering::get_history(const Board &b, int ply, int start, int dest)
{
   if (b.Side() == White)
      return HistoryW[start][dest];
   else
      return HistoryB[start][dest];
}

Move Move_Ordering::get_killer(int ply,int index)
{
   return (index == 0) ? Killers1[ply] : Killers2[ply];
}

void Move_Ordering::update_killers()
{
   // not currently implemented; we just clear the killer lists
   clear_killer();
}


