// Copyright 1994-1997 by Jon Dart.  All Rights Reserved.

#ifndef _SEARCH_H
#define _SEARCH_H

#include "board.h"
#include "move.h"
#include "movegen.h"
#include "hash.h"
#include "stats.h"
#include <time.h>


class Search
{
    // Encapsulates alpha-beta search routine.

    public:
        
    enum Search_Type { Fixed_Depth, Time_Limit, Fixed_Time };

    enum Extension_Type { RECAPTURE=1, CHECK=2, PAWN_PUSH=4, CAPTURE=8};

    // There are 4 levels of verbosity.  Silent mode does no output to
    // the console - it is used by the Windows GUI version.  Debug
    // level is used to output debug info.  Whisper is used to "whisper"
    // comments on a chess server.  "Trace" is used in trace mode
    // (-t) of the chess server client.
    enum Talk_Level { Silent, Debug, Whisper, Trace };
    
    Search();
    
    ~Search();
    
    void terminate_now(int makeMove = 0);
    
    Move find_best_move( 
            const Board &board, 
            Search::Search_Type srcType,
            int32 time_limit,
            int32 xtratime,
            int ply_limit,
            const int background,
            Statistics &stats,
            Talk_Level t = Silent);
    
    int get_ply() const;
    // return current depth of search.

    // return # of nodes visited
    unsigned long get_numnodes() const;
            
    // Get the starting time for the search in centiseconds
    time_t get_start_time() const
    {
        return start_time;
    }
    
    Search_Type get_search_type() const
    {
        return type_of_search;
    }
    
    void set_time_limit(time_t limit,time_t xtra_time_limit)
    {
       type_of_search = Time_Limit;
       time_target = limit;
       xtra_time = xtra_time_limit;
    }
    
    void set_ply_limit(int ply)
    {
       type_of_search = Fixed_Depth;
       ply_limit = ply;
    }
    
    time_t get_time_limit() const
    { 
        return time_target + (time_added ? xtra_time : 0);
    }
    
    void set_talk_level(Talk_Level t)
    {
       talkLevel = t;
    }
    
    ColorType get_side_to_move() const
    {
        return side_to_move;
    }
    
    int is_background_search() const
    {
        return bkgrnd_search;
    }
    
    void set_background(int b);
    
    void set_time_up( const int b )
    {
        time_up = b;
    }
    
    int was_terminated() const
    {
        return terminated;
    }
    
     void register_command_function(void (*cmd)(char *,int &))
     {
        command_function = cmd;
     }
     
     void register_post_function(void (*post)(const Statistics &))
     {
        post_function = post;
     }
     
     void register_pending_function(void (*cmd)(const char *))
     {
        pending_function = cmd;
     }
     
     void register_terminate_function(int (*term)(const Statistics &))
     {
        terminate_function = term;
     }
     
    // Calculate the recommended hash table size (expressed
    // in number of entries).  
    static long calc_hash_size();
    
    int is_draw( const Board &board, int &rep_count, int ply );
    
    int draw_score(const Board & board);
    
    void reinit();

    void update_stats(int alpha, int beta);

    // output something to the console
    void show_status(const Board &board, const char ch);

    // check console input
    int check_input(const Board &);

    void setRatingDiff(int rdiff);

    time_t getTimeAdded() const {
        return time_added;
    }

private:
        void iterate(const Board &board);
        void init(const Board &);
        void cleanup(int clearHash);
        unsigned generate_moves(
               Board & board, Move_Generator & mg,
               const unsigned ply,
               Move * moves);
        int evalu8(const Board & board, int alpha, int beta, int ply);
        int ply0_move_search(Board & board, int alpha, int beta,
            int depth, int &terminate);
        int move_search(Board & board, int alpha, int beta,
            const int ply, int depth, int in_pv, int &terminate);
        int quiesce(Board & board, int alpha, int beta,
            const int ply, int depth);
        int check_time(const Board &,int ply, int rank);
        inline int NullOk(const Board &board,int in_pv,int in_check,int ply,
			  int fprune,int depth);
        
        // make this class uncopyable:
        Search( const Search & );
        Search &operator = ( const Search & );

        time_t time_target, xtra_time, time_added, start_time;
        Search_Type type_of_search;
        int ply_limit;
        int bkgrnd_search;
        unsigned long num_nodes, num_qnodes, num_moves;
        Move *ply0moves;
        int ply0move_count;
        int *ply0scores;
        int iteration_depth;
        int current_score; // The score displayed to the user
        byte age;
        long hash_hits, hash_inserts, hash_searches,
          hash_replaces, hash_inserts_failed;
        int time_up;
        int terminated;
        int easy;
        Move easy_move;
        int time_check_counter;
        Board initial_board;
        Statistics *stats;

        ColorType side_to_move;
        StateType state;
        Move *moves_generated[Constants::MaxPly];
        Move best[Constants::MaxPly];
        Move last_move[Constants::MaxPly];
        Move best_moves[Constants::MaxPly];
        int best_scores[Constants::MaxPly];
        int extensions[Constants::MaxPly];
        int checks[Constants::MaxPly];
        Move ply0first; // first move to search
        int root_total_mat;
        // search parameters
        int null_depth;
        Talk_Level talkLevel;
        int pipe;

        // pointer to function, called to process a command
        // received during the search (console programs only).
        void (*command_function)(char *,int &);

        // pointer to function, called to output status during
        // a search (used in console mode only).
        void (*post_function)(const Statistics &);

        int (*terminate_function)(const Statistics &);
        
        // Function to add a command to the deferred execution
        // stack (UNIX_ONLY)
        void (*pending_function)(const char*);
        
        int calc_extensions(const Board &board,int ply,int depth,int in_check,int forced,int threat);
        
        int hash_init_done;
        int use_hash_table;
        int hash_table_size;
        int ratingDiff;
        int ratingFactor; 
};

#endif

